# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from osc_lib.tests import utils as osc_tests_utils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import function_alias
from qinlingclient.tests.unit.osc.v1 import fakes


class TestFunctionAlias(fakes.TestQinlingClient):

    def setUp(self):
        super(TestFunctionAlias, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.FUNCTION_ALIAS_COLUMNS
        self.data = []

        aliases = fakes.FakeFunctionAlias.create_function_aliases(count=3)
        self._function_aliases = aliases
        for a in self._function_aliases:
            self.data.append((a.name, a.function_id, a.description,
                              a.function_version, a.project_id,
                              a.created_at, a.updated_at))


class TestListFunctionAlias(TestFunctionAlias):

    def setUp(self):
        super(TestListFunctionAlias, self).setUp()
        self.cmd = function_alias.List(self.app, None)

        self.columns = [c.capitalize() for c in base.FUNCTION_ALIAS_COLUMNS]

        self.client.function_aliases.list = mock.Mock(
            return_value=self._function_aliases
        )

    def test_function_alias_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_aliases.list.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_function_alias_list_with_filter(self):
        arglist = ['--filter', 'name=has:alias',
                   '--filter', 'function_id=has:900dcafe']
        verifylist = [
            ('filters', ['name=has:alias', 'function_id=has:900dcafe']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_aliases.list.assert_called_once_with(
            name='has:alias', function_id='has:900dcafe'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_function_alias_list_with_invalid_filter(self):
        arglist = ['--filter', 'function_version']
        verifylist = [
            ('filters', ['function_version']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            ValueError,
            '^Invalid filter: function_version$',
            self.cmd.take_action, parsed_args
        )


class TestCreateFunctionAlias(TestFunctionAlias):

    def setUp(self):
        super(TestCreateFunctionAlias, self).setUp()
        self.cmd = function_alias.Create(self.app, None)

    def _create_fake_function_alias(self, attrs=None):
        # Allow to fake different create results
        a = fakes.FakeFunctionAlias.create_one_function_alias(attrs)
        self.client.function_aliases.create = mock.Mock(return_value=a)
        data = (a.name, a.function_id, a.description, a.function_version,
                a.project_id, a.created_at, a.updated_at)
        return data

    def test_function_alias_create_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_alias_create_required_options(self):
        """Create a function alias.

        1. use function_id,
        2. all other params except the required ones are not set.
        """
        alias_name = 'FAKE_ALIAS_NAME'
        function_id = self._function_aliases[0].function_id
        attrs = {'name': alias_name, 'function_id': function_id}
        created_data = self._create_fake_function_alias(attrs)

        arglist = [alias_name, '--function', function_id]
        verifylist = [
            ('name', alias_name),
            ('function', function_id),
            ('function_version', 0),
            ('description', ''),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_aliases.create.assert_called_once_with(
            alias_name,
            **{'function_id': function_id,
               'function_version': 0,
               'description': ''}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_function_alias_create_all_options(self):
        """Create a function alias.

        1. use function name to find the function_id,
        2. all optional params are specified.
        """
        alias_name = 'FAKE_ALIAS_NAME'
        function = fakes.FakeFunction.create_one_function()
        function_name = function.name
        function_id = function.id
        function_version = 1
        alias_description = 'This is a newly created function alias.'
        attrs = {'name': alias_name,
                 'function_id': function_id,
                 'function_version': function_version,
                 'description': alias_description}
        created_data = self._create_fake_function_alias(attrs)

        # Use to find the function id with its name
        self.client.functions.find.return_value = function

        arglist = [alias_name,
                   '--function', function_name,
                   '--function-version', str(function_version),
                   '--description', alias_description]
        verifylist = [
            ('name', alias_name),
            ('function', function_name),
            ('function_version', function_version),
            ('description', alias_description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_aliases.create.assert_called_once_with(
            alias_name,
            **{'function_id': function_id,
               'function_version': function_version,
               'description': alias_description}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.functions.find.assert_called_once_with(name=function_name)

    def test_function_alias_create_version_not_integer(self):
        # function_version should be an integer value
        alias_name = 'FAKE_ALIAS_NAME'
        function_id = self._function_aliases[0].function_id

        arglist = [alias_name,
                   '--function', function_id,
                   '--function-version', 'NOT_A_INTEGER']
        verifylist = [
            ('name', alias_name),
            ('function', function_id),
            ('function_version', 0),
            ('description', None),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestDeleteFunctionAlias(TestFunctionAlias):

    def setUp(self):
        super(TestDeleteFunctionAlias, self).setUp()
        self.cmd = function_alias.Delete(self.app, None)
        self.client.function_aliases.delete = mock.Mock(return_value=None)

    def test_function_alias_delete_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_alias_delete_one(self):
        alias_name = self._function_aliases[0].name
        arglist = [alias_name]
        verifylist = [('name', [alias_name])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_aliases.delete.assert_called_once_with(alias_name)

    def test_function_alias_delete_multiple(self):
        alias_names = [a.name for a in self._function_aliases]
        arglist = alias_names
        verifylist = [('name', alias_names)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        calls = [mock.call(a_name) for a_name in alias_names]
        self.assertEqual(len(alias_names),
                         self.client.function_aliases.delete.call_count)
        self.client.function_aliases.delete.assert_has_calls(calls)

    def test_function_alias_delete_multiple_exception(self):
        alias_names = [a.name for a in self._function_aliases]
        arglist = alias_names
        verifylist = [('name', alias_names)]

        self.client.function_aliases.delete = mock.Mock(side_effect=[
            None, RuntimeError, None
        ])

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to delete the specified function_alias\(s\)\.$',
            self.cmd.take_action, parsed_args)

        # The second deleteion failed, but the third is done normally
        calls = [mock.call(a_name) for a_name in alias_names]
        self.assertEqual(len(alias_names),
                         self.client.function_aliases.delete.call_count)
        self.client.function_aliases.delete.assert_has_calls(calls)


class TestShowFunctionAlias(TestFunctionAlias):

    def setUp(self):
        super(TestShowFunctionAlias, self).setUp()
        self.cmd = function_alias.Show(self.app, None)
        self.client.function_aliases.get = mock.Mock(
            return_value=self._function_aliases[0])

    def test_function_alias_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_alias_show(self):
        alias_name = self._function_aliases[0].name
        arglist = [alias_name]
        verifylist = [('name', alias_name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_aliases.get.assert_called_once_with(alias_name)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)


class TestUpdateFunctionAlias(TestFunctionAlias):

    def setUp(self):
        super(TestUpdateFunctionAlias, self).setUp()
        self.cmd = function_alias.Update(self.app, None)

    def _update_fake_function_alias(self, attrs=None):
        # Allow to fake different update results
        a = fakes.FakeFunctionAlias.create_one_function_alias(attrs)
        self.client.function_aliases.update = mock.Mock(return_value=a)
        data = (a.name, a.function_id, a.description, a.function_version,
                a.project_id, a.created_at, a.updated_at)
        return data

    def test_function_alias_update_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_alias_update_required_options(self):
        """Update a function alias.

        Do nothing as only the alias name is specified.
        """
        alias_name = self._function_aliases[0].name
        attrs = {'name': alias_name}
        updated_data = self._update_fake_function_alias(attrs)

        arglist = [alias_name]
        verifylist = [
            ('name', alias_name),
            ('function', None),
            ('function_version', None),
            ('description', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_aliases.update.assert_called_once_with(
            alias_name,
            **{'function_id': None,
               'function_version': None,
               'description': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_function_alias_update_all_options(self):
        """Update a function alias.

        use function name to find the function_id,
        """
        alias_name = self._function_aliases[0].name
        function = fakes.FakeFunction.create_one_function()
        function_name = function.name
        function_id = function.id
        function_version = 1
        alias_description = 'This is a updated function alias.'
        attrs = {'name': alias_name,
                 'function_id': function_id,
                 'function_version': function_version,
                 'description': alias_description}
        created_data = self._update_fake_function_alias(attrs)

        # Use to find the function id with its name
        self.client.functions.find.return_value = function

        arglist = [alias_name,
                   '--function', function_name,
                   '--function-version', str(function_version),
                   '--description', alias_description]
        verifylist = [
            ('name', alias_name),
            ('function', function_name),
            ('function_version', str(function_version)),
            ('description', alias_description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_aliases.update.assert_called_once_with(
            alias_name,
            **{'function_id': function_id,
               'function_version': str(function_version),
               'description': alias_description}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.functions.find.assert_called_once_with(name=function_name)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid

from oslo_serialization import jsonutils

from qinlingclient.common import exceptions
from qinlingclient.tests.unit.v1 import test_client

JOB_1 = {'id': str(uuid.uuid4()), 'name': 'job_1',
         'function_id': str(uuid.uuid4())}
JOB_2 = {'id': str(uuid.uuid4()), 'name': 'job_2',
         'function_id': JOB_1['function_id']}
JOB_3 = {
    'id': str(uuid.uuid4()),
    'name': 'job_2',
    'function_alias': 'alias_1',
}

LIST_JOBS_RESP = {
    'jobs': [JOB_1, JOB_2]
}


class TestJob(test_client.TestQinlingClient):

    _error_message = "Test error message."

    def test_list_job(self):
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/jobs',
            headers={'Content-Type': 'application/json'},
            json=LIST_JOBS_RESP,
            status_code=200
        )
        ret = self.client.jobs.list()
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [JOB_1, JOB_2],
            [resource.to_dict() for resource in ret])

    def test_create_job(self):
        function_id = JOB_1['function_id']
        request_data = {'function_alias': None, 'function_id': function_id,
                        'function_version': 0,
                        'name': None, 'first_execution_time': None,
                        'pattern': None, 'function_input': None,
                        'count': None}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/jobs',
            headers={'Content-Type': 'application/json'},
            json=JOB_1,
            status_code=201
        )

        ret = self.client.jobs.create(function_id=function_id)

        self.assertEqual(JOB_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_job_all_options(self):
        function_id = JOB_1['function_id']
        function_version = 1
        name = JOB_1['name']
        first_execution_time = '2018-08-16T08:00:00'
        pattern = '0 * * * *'
        function_input = '{"name": "Qinling"}'
        count = 3
        request_data = {'function_alias': None,
                        'function_id': function_id,
                        'function_version': function_version,
                        'name': name,
                        'first_execution_time': first_execution_time,
                        'pattern': pattern, 'function_input': function_input,
                        'count': count}

        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/jobs',
            headers={'Content-Type': 'application/json'},
            json=JOB_1,
            status_code=201
        )

        ret = self.client.jobs.create(
            function_id=function_id, function_version=function_version,
            name=name,
            first_execution_time=first_execution_time, pattern=pattern,
            function_input=function_input, count=count)

        self.assertEqual(JOB_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_job_error(self):
        function_id = JOB_1['function_id']
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/jobs',
            headers={'Content-Type': 'application/json'},
            text='{"faultstring": "%s"}' % self._error_message,
            status_code=400
        )
        self.assertRaisesRegex(
            exceptions.HTTPBadRequest,
            self._error_message,
            self.client.jobs.create,
            function_id)

    def test_delete_job(self):
        job_id = JOB_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/jobs/%s' % job_id,
            status_code=204
        )
        ret = self.client.jobs.delete(job_id)
        self.assertIsNone(ret)

    def test_delete_job_error(self):
        job_id = JOB_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/jobs/%s' % job_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.jobs.delete,
            job_id)

    def test_get_job(self):
        job_id = JOB_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/jobs/%s' % job_id,
            headers={'Content-Type': 'application/json'},
            json=JOB_2,
            status_code=200
        )
        ret = self.client.jobs.get(job_id)
        self.assertEqual(JOB_2, ret.to_dict())

    def test_get_job_error(self):
        job_id = JOB_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/jobs/%s' % job_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.jobs.get,
            job_id)

    def test_update_job(self):
        job_id = JOB_2['id']
        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/jobs/%s' % job_id,
            headers={'Content-Type': 'application/json'},
            json=JOB_2,
            status_code=200
        )
        ret = self.client.jobs.update(job_id)
        self.assertEqual(JOB_2, ret.to_dict())

    def test_update_job_with_params(self):
        job_id = JOB_2['id']
        name = 'renamed_job'
        pattern = '0 1 * * *'
        request_data = {'name': name, 'pattern': pattern}

        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/jobs/%s' % job_id,
            headers={'Content-Type': 'application/json'},
            json=JOB_2,
            status_code=200
        )
        ret = self.client.jobs.update(job_id, name=name, pattern=pattern)
        self.assertEqual(JOB_2, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_update_job_error(self):
        job_id = JOB_2['id']
        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/jobs/%s' % job_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=400
        )
        self.assertRaisesRegex(
            exceptions.HTTPBadRequest,
            self._error_message,
            self.client.jobs.update,
            job_id)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid

from oslo_serialization import jsonutils

from qinlingclient.common import exceptions
from qinlingclient.tests.unit.v1 import test_client

RUNTIME_1 = {'id': str(uuid.uuid4()), 'name': 'runtime_1'}
RUNTIME_2 = {'id': str(uuid.uuid4()), 'name': 'runtime_2'}

LIST_RUNTIMES_RESP = {
    'runtimes': [RUNTIME_1, RUNTIME_2]
}

RUNTIME_POOL = {'name': RUNTIME_2['id'],
                'capacity': {'available': 5, 'total': 5}}


class TestRuntime(test_client.TestQinlingClient):

    _error_message = "Test error message."

    def test_list_runtime(self):
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/runtimes',
            headers={'Content-Type': 'application/json'},
            json=LIST_RUNTIMES_RESP,
            status_code=200
        )
        ret = self.client.runtimes.list()
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [RUNTIME_1, RUNTIME_2],
            [resource.to_dict() for resource in ret])

    def test_create_runtime(self):
        image_name = 'image_name'
        request_data = {'image': image_name, 'trusted': True,
                        'is_public': True}

        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/runtimes',
            headers={'Content-Type': 'application/json'},
            json=RUNTIME_1,
            status_code=201
        )
        ret = self.client.runtimes.create(image_name)

        self.assertEqual(RUNTIME_1, ret.to_dict())
        self.assertEqual(request_data,
                         jsonutils.loads(self.requests_mock.last_request.text))

    def test_create_runtime_all_options(self):
        image_name = 'image_name'
        runtime_name = 'runtime_name'
        description = 'A newly created runtime.'
        trusted = False
        is_public = False

        request_data = {'image': image_name, 'trusted': trusted,
                        'name': runtime_name, 'is_public': is_public,
                        'description': description}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/runtimes',
            headers={'Content-Type': 'application/json'},
            json=RUNTIME_1,
            status_code=201
        )

        ret = self.client.runtimes.create(
            image_name, name=runtime_name, description=description,
            trusted=False, is_public = False,
        )

        self.assertEqual(RUNTIME_1, ret.to_dict())
        self.assertEqual(request_data,
                         jsonutils.loads(self.requests_mock.last_request.text))

    def test_create_runtime_error(self):
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/runtimes',
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=400
        )
        self.assertRaisesRegex(
            exceptions.HTTPBadRequest,
            self._error_message,
            self.client.runtimes.create,
            'image_name'
        )

    def test_delete_runtime(self):
        runtime_id = RUNTIME_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/runtimes/%s' % runtime_id,
            status_code=204
        )
        ret = self.client.runtimes.delete(runtime_id)
        self.assertIsNone(ret)

    def test_delete_runtime_error(self):
        runtime_id = RUNTIME_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/runtimes/%s' % runtime_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=403
        )
        self.assertRaisesRegex(
            exceptions.HTTPForbidden,
            self._error_message,
            self.client.runtimes.delete,
            runtime_id
        )

    def test_get_runtime(self):
        runtime_id = RUNTIME_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/runtimes/%s' % runtime_id,
            headers={'Content-Type': 'application/json'},
            json=RUNTIME_2,
            status_code=200
        )
        ret = self.client.runtimes.get(runtime_id)
        self.assertEqual(RUNTIME_2, ret.to_dict())

    def test_get_runtime_error(self):
        runtime_id = RUNTIME_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/runtimes/%s' % runtime_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.runtimes.get,
            runtime_id
        )

    def test_get_pool_runtime(self):
        runtime_id = RUNTIME_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/runtimes/%s/pool' % runtime_id,
            headers={'Content-Type': 'application/json'},
            json=RUNTIME_POOL,
            status_code=200
        )
        ret = self.client.runtimes.get_pool(runtime_id)
        self.assertEqual(RUNTIME_POOL, ret.to_dict())

    def test_get_pool_runtime_error(self):
        runtime_id = RUNTIME_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/runtimes/%s/pool' % runtime_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.runtimes.get_pool,
            runtime_id
        )



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import six
from six.moves.urllib.parse import urlencode
import uuid

from oslo_serialization import jsonutils

from qinlingclient.common import exceptions
from qinlingclient.tests.unit.v1 import test_client

FUNCTION_1 = {'id': str(uuid.uuid4()), 'name': 'function_1'}
FUNCTION_2 = {'id': str(uuid.uuid4()), 'name': 'function_2'}

LIST_FUNCTIONS_RESP = {
    'functions': [FUNCTION_1, FUNCTION_2]
}


class TestFunction(test_client.TestQinlingClient):

    _error_message = "Test error message."

    def test_list_function(self):
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/functions',
            headers={'Content-Type': 'application/json'},
            json=LIST_FUNCTIONS_RESP,
            status_code=200
        )
        ret = self.client.functions.list()
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [FUNCTION_1, FUNCTION_2],
            [resource.to_dict() for resource in ret])

    def test_list_function_with_params(self):
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/functions?name=test&count=0',
            headers={'Content-Type': 'application/json'},
            json=LIST_FUNCTIONS_RESP,
            status_code=200
        )
        ret = self.client.functions.list(name='test', count=0)
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [FUNCTION_1, FUNCTION_2],
            [resource.to_dict() for resource in ret])

    def test_create_function(self):
        runtime_id = 'runtime_id'
        code = {'source': 'package', 'md5sum': 'MD5SUM'}
        data = {'runtime_id': runtime_id, 'code': jsonutils.dumps(code)}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/functions',
            headers={'Content-Type': 'application/json'},
            json=FUNCTION_1,
            status_code=201
        )
        ret = self.client.functions.create(code, runtime=runtime_id)
        self.assertEqual(FUNCTION_1, ret.to_dict())
        self.assertEqual(urlencode(data), self.requests_mock.last_request.text)

    def test_create_function_all_options(self):
        runtime_id = 'runtime_id'
        code = {'source': 'package', 'md5sum': 'MD5SUM'}
        package_content = 'package file content'
        package = six.StringIO(package_content)
        cpu = '100'
        memory_size = '33554432'
        data = {'runtime_id': runtime_id, 'code': jsonutils.dumps(code),
                'cpu': cpu, 'memory_size': memory_size}

        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/functions',
            headers={'Content-Type': 'application/json'},
            json=FUNCTION_1,
            status_code=201
        )
        ret = self.client.functions.create(
            code, runtime=runtime_id, package=package,
            cpu=cpu, memory_size=memory_size)
        self.assertEqual(FUNCTION_1, ret.to_dict())

        # Request body is a multipart/form-data
        request_body = self.requests_mock.last_request.body.decode('utf-8')
        param_base_str = ('Content-Disposition: form-data; name="{key}"'
                          '\r\n\r\n{value}')
        file_base_str = ('Content-Disposition: form-data; name="{name}"; '
                         'filename="{filename}"\r\n\r\n{content}')
        for k, v in data.items():
            param_str = param_base_str.format(key=k, value=v)
            self.assertIn(param_str, request_body)
        # filename is same as name since we use a StringIO instead of a
        # real file object in this test.
        file_str = file_base_str.format(name='package', filename='package',
                                        content=package_content)
        self.assertIn(file_str, request_body)

    def test_create_function_error(self):
        runtime_id = 'runtime_id'
        code = {'source': 'package', 'md5sum': 'MD5SUM'}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/functions',
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=400
        )
        self.assertRaisesRegex(
            exceptions.HTTPBadRequest,
            self._error_message,
            self.client.functions.create,
            code, runtime=runtime_id)

    def test_delete_function(self):
        function_id = FUNCTION_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/functions/%s' % function_id,
            status_code=204
        )
        ret = self.client.functions.delete(function_id)
        self.assertIsNone(ret)

    def test_delete_function_error(self):
        function_id = FUNCTION_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/functions/%s' % function_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=403
        )
        self.assertRaisesRegex(
            exceptions.HTTPForbidden,
            self._error_message,
            self.client.functions.delete,
            function_id
        )

    def test_get_function(self):
        function_id = FUNCTION_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/functions/%s' % function_id,
            headers={'Content-Type': 'application/json'},
            json=FUNCTION_2,
            status_code=200
        )
        ret = self.client.functions.get(function_id)
        self.assertEqual(FUNCTION_2, ret.to_dict())
        self.assertFalse(self.requests_mock.last_request.stream)

    def test_get_function_download(self):
        function_id = FUNCTION_2['id']
        function_data = 'function package data'
        self.requests_mock.register_uri(
            'GET',
            (test_client.QINLING_URL +
             '/v1/functions/%s?download=true' % function_id),
            headers={'Content-Type': 'application/zip'},
            text=function_data,
            status_code=200
        )
        ret = self.client.functions.get(function_id, download=True)
        self.assertEqual(function_data, ret.text)
        self.assertEqual('application/zip', ret.headers['content-type'])
        self.assertTrue(self.requests_mock.last_request.stream)

    def test_get_function_error(self):
        function_id = FUNCTION_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/functions/%s' % function_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.functions.get,
            function_id
        )

    def test_update_function(self):
        function_id = FUNCTION_2['id']
        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/functions/%s' % function_id,
            headers={'Content-Type': 'application/json'},
            json=FUNCTION_2,
            status_code=200
        )
        ret = self.client.functions.update(function_id)
        self.assertEqual(FUNCTION_2, ret.to_dict())

    def test_update_function_all_options(self):
        function_id = FUNCTION_2['id']
        code = {'source': 'package'}
        package_content = 'updated package file content'
        package = six.StringIO(package_content)
        cpu = '100'
        memory_size = '33554432'
        data = {'source': 'package', 'cpu': cpu, 'memory_size': memory_size}

        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/functions/%s' % function_id,
            headers={'Content-Type': 'application/json'},
            json=FUNCTION_2,
            status_code=200
        )
        ret = self.client.functions.update(
            function_id, code=code, package=package,
            cpu=cpu, memory_size=memory_size)
        self.assertEqual(FUNCTION_2, ret.to_dict())

        # Request body is a multipart/form-data
        request_body = self.requests_mock.last_request.body.decode('utf-8')
        param_base_str = ('Content-Disposition: form-data; name="{key}"'
                          '\r\n\r\n{value}')
        file_base_str = ('Content-Disposition: form-data; name="{name}"; '
                         'filename="{filename}"\r\n\r\n{content}')
        for k, v in data.items():
            param_str = param_base_str.format(key=k, value=v)
            self.assertIn(param_str, request_body)
        # filename is same as name since we use a StringIO instead of a
        # real file object in this test.
        file_str = file_base_str.format(name='package', filename='package',
                                        content=package_content)
        self.assertIn(file_str, request_body)

    def test_update_function_error(self):
        function_id = FUNCTION_2['id']
        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/functions/%s' % function_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=403
        )
        self.assertRaisesRegex(
            exceptions.HTTPForbidden,
            self._error_message,
            self.client.functions.update,
            function_id)

    def test_detach_function(self):
        function_id = FUNCTION_2['id']
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/functions/%s/detach' % function_id,
            status_code=202
        )
        ret = self.client.functions.detach(function_id)
        self.assertEqual('', ret.text)
        self.assertEqual(202, ret.status_code)

    def test_detach_function_error(self):
        function_id = FUNCTION_2['id']
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/functions/%s/detach' % function_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.functions.detach,
            function_id)

    def test_scaleup_function(self):
        function_id = FUNCTION_1['id']
        self.requests_mock.register_uri(
            'POST',
            (test_client.QINLING_URL +
             '/v1/functions/%s/scale_up' % function_id),
            status_code=202
        )
        resp, text = self.client.functions.scaleup(function_id)
        self.assertEqual('', text)
        self.assertEqual(202, resp.status_code)
        self.assertEqual(jsonutils.dumps({'count': 1}),
                         self.requests_mock.last_request.text)

    def test_scaleup_function_error(self):
        function_id = FUNCTION_1['id']
        self.requests_mock.register_uri(
            'POST',
            (test_client.QINLING_URL +
             '/v1/functions/%s/scale_up' % function_id),
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.functions.scaleup,
            function_id)

    def test_scaledown_function(self):
        function_id = FUNCTION_1['id']
        self.requests_mock.register_uri(
            'POST',
            (test_client.QINLING_URL +
             '/v1/functions/%s/scale_down' % function_id),
            status_code=202
        )
        resp, text = self.client.functions.scaledown(function_id, count=2)
        self.assertEqual('', text)
        self.assertEqual(202, resp.status_code)
        self.assertEqual(jsonutils.dumps({'count': 2}),
                         self.requests_mock.last_request.text)

    def test_scaledown_function_error(self):
        function_id = FUNCTION_1['id']
        self.requests_mock.register_uri(
            'POST',
            (test_client.QINLING_URL +
             '/v1/functions/%s/scale_down' % function_id),
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.functions.scaledown,
            function_id)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from osc_lib.command import command
from osc_lib import utils as osc_utils
from oslo_utils import uuidutils

from qinlingclient.osc.v1 import base
from qinlingclient import utils as q_utils


class List(base.QinlingLister):
    columns = base.JOB_COLUMNS
    filtered_columns = base.FILTERED_JOB_COLUMNS

    def _get_resources(self, parsed_args):
        client = self.app.client_manager.function_engine

        return client.jobs.list(**base.get_filters(parsed_args))


class Create(command.ShowOne):
    columns = base.JOB_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            "--function",
            help="Function name or ID.",
        )
        parser.add_argument(
            "--function-version",
            type=int,
            default=0,
            help="Function version number. Default: 0",
        )
        parser.add_argument(
            "--function-alias",
            help="Function alias which corresponds to a specific function and "
                 "version. When function alias is specified, function and "
                 "function version are not needed.",
        )
        parser.add_argument(
            "--name",
            help="Job name."
        )
        parser.add_argument(
            "--first-execution-time",
            help="The earliest execution time(UTC) for the job."
        )
        parser.add_argument(
            "--pattern",
            help="The cron pattern for job execution."
        )
        parser.add_argument(
            "--function-input",
            help="Function input."
        )
        parser.add_argument(
            "--count",
            type=int,
            help="Expected number of executions triggered by job."
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        function_alias = parsed_args.function_alias
        if function_alias:
            function_id = None
            function_version = None
        else:
            function_version = parsed_args.function_version
            function_id = parsed_args.function
            if not uuidutils.is_uuid_like(function_id):
                # Try to find the function id with name
                function_id = q_utils.find_resource_id_by_name(
                    client.functions, function_id)

        job = client.jobs.create(
            function_alias=function_alias,
            function_id=function_id,
            function_version=function_version,
            name=parsed_args.name,
            first_execution_time=parsed_args.first_execution_time,
            pattern=parsed_args.pattern,
            function_input=parsed_args.function_input,
            count=parsed_args.count
        )

        return self.columns, osc_utils.get_item_properties(job, self.columns)


class Delete(base.QinlingDeleter):
    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'job',
            nargs='+',
            help='Job ID(s).'
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        self.delete = client.jobs.delete
        self.resource = 'job'

        self.delete_resources(parsed_args.job)


class Show(command.ShowOne):
    columns = base.JOB_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Show, self).get_parser(prog_name)
        parser.add_argument('job', help='Job ID.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        job = client.jobs.get(parsed_args.job)

        return self.columns, osc_utils.get_item_properties(job, self.columns)


class Update(command.ShowOne):
    columns = base.JOB_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'id',
            help='Job ID.'
        )
        parser.add_argument(
            "--name",
            help="Job name."
        )
        parser.add_argument(
            "--status",
            choices=['running', 'paused', 'done', 'cancelled'],
            help="Job status."
        )
        parser.add_argument(
            "--next-execution-time",
            help="The next execution time(UTC) for the job."
        )
        parser.add_argument(
            "--pattern",
            help="The cron pattern for job execution."
        )
        parser.add_argument(
            "--function-input",
            help="Function input."
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        job = client.jobs.update(
            parsed_args.id,
            name=parsed_args.name,
            status=parsed_args.status,
            pattern=parsed_args.pattern,
            next_execution_time=parsed_args.next_execution_time,
            function_input=parsed_args.function_input,
        )

        return self.columns, osc_utils.get_item_properties(job, self.columns)



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.common import base

URL_TEMPLATE = "/v1/aliases"


class FunctionAlias(base.Resource):
    pass


class FunctionAliasManager(base.Manager):
    resource_class = FunctionAlias

    def list(self, **kwargs):
        url = URL_TEMPLATE
        return self._list(url, response_key='function_aliases')

    def create(self, name, function_id, function_version=0, description=""):
        url = URL_TEMPLATE
        request_data = {
            'name': name,
            'function_id': function_id,
            'function_version': function_version,
            'description': description
        }

        return self._create(url, data=request_data)

    def delete(self, name):
        url = URL_TEMPLATE + '/%s' % name
        self._delete(url)

    def get(self, name):
        url = URL_TEMPLATE + '/%s' % name
        return self._get(url)

    def update(self, name, function_id=None, function_version=None,
               description=None):
        url = URL_TEMPLATE + '/%s' % name
        request_data = {
            'function_id': function_id,
            'function_version': function_version,
            'description': description
        }

        return self._update(url, data=request_data)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

import six

AUTH_TOKEN = 'foobar'
AUTH_URL = 'http://0.0.0.0'


class FakeResource(object):

    def __init__(self, manager=None, info=None, loaded=False, methods=None):
        """Set attributes and methods for a resource.

        :param manager:
            The resource manager
        :param Dictionary info:
            A dictionary with all attributes
        :param bool loaded:
            True if the resource is loaded in memory
        :param Dictionary methods:
            A dictionary with all methods
        """
        info = info or {}
        methods = methods or {}

        self.__name__ = type(self).__name__
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._add_methods(methods)
        self._loaded = loaded

    def _add_details(self, info):
        for (k, v) in six.iteritems(info):
            setattr(self, k, v)

    def _add_methods(self, methods):
        """Fake methods with MagicMock objects.

        For each <@key, @value> pairs in methods, add an callable MagicMock
        object named @key as an attribute, and set the mock's return_value to
        @value. When users access the attribute with (), @value will be
        returned, which looks like a function call.
        """
        for (name, ret) in six.iteritems(methods):
            method = mock.Mock(return_value=ret)
            setattr(self, name, method)

    def __repr__(self):
        reprkeys = sorted(k for k in self.__dict__.keys() if k[0] != '_' and
                          k != 'manager')
        info = ", ".join("%s=%s" % (k, getattr(self, k)) for k in reprkeys)
        return "<%s %s>" % (self.__class__.__name__, info)

    def keys(self):
        return self._info.keys()

    def to_dict(self):
        return self._info

    @property
    def info(self):
        return self._info

    def __getitem__(self, item):
        return self._info.get(item)

    def get(self, item, default=None):
        return self._info.get(item, default)

    def pop(self, key, default_value=None):
        return self.info.pop(key, default_value)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""OpenStackClient plugin for Function service."""

from osc_lib import utils
from oslo_log import log as logging

from qinlingclient.i18n import _

LOG = logging.getLogger(__name__)

DEFAULT_FUNCTION_ENGINE_API_VERSION = "1"
API_NAME = "function_engine"
API_VERSION_OPTION = "os_function_engine_api_version"
API_VERSIONS = {
    '1': 'qinlingclient.v1.client.Client',
}


def make_client(instance):
    """Returns an qinling service client"""
    function_engine_client = utils.get_client_class(
        API_NAME,
        instance._api_version[API_NAME],
        API_VERSIONS)

    LOG.debug("Instantiating function-engine client: {0}".format(
        function_engine_client))

    kwargs = {
        'session': instance.session,
        'service_type': 'function-engine',
        'region_name': instance._region_name
    }

    qinling_endpoint = instance.get_configuration().get('qinling_url')
    if not qinling_endpoint:
        qinling_endpoint = instance.get_endpoint_for_service_type(
            'function-engine',
            region_name=instance._region_name,
            interface=instance._interface
        )

    client = function_engine_client(qinling_endpoint, **kwargs)
    return client


def build_option_parser(parser):
    """Hook to add global options"""
    parser.add_argument(
        '--os-function-engine-api-version',
        metavar='<function-engine-api-version>',
        default=utils.env(
            'OS_FUNCTION_ENGINE_API_VERSION',
            default=DEFAULT_FUNCTION_ENGINE_API_VERSION
        ),
        help=_(
            "Function engine API version, default={0}"
            "(Env:OS_FUNCTION_ENGINE_API_VERSION)").format(
            DEFAULT_FUNCTION_ENGINE_API_VERSION
        )
    )
    parser.add_argument('--qinling-url',
                        default=utils.env('QINLING_URL'),
                        help=_('Defaults to env[QINLING_URL].'))
    return parser



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.common import base


class FunctionWorker(base.Resource):
    pass


class WorkerManager(base.Manager):
    resource_class = FunctionWorker

    def list(self, function_id):
        url = '/v1/functions/%s/workers' % function_id
        return self._list(url, response_key='workers')



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid

from oslo_serialization import jsonutils

from qinlingclient.common import exceptions
from qinlingclient.tests.unit.v1 import test_client

VERSION_1 = {'id': str(uuid.uuid4()), 'function_id': str(uuid.uuid4()),
             'version_number': 1}
VERSION_2 = {'id': str(uuid.uuid4()), 'function_id': VERSION_1['function_id'],
             'version_number': 2}

LIST_FUNCTION_VERSIONS_RESP = {
    'function_versions': [VERSION_1, VERSION_2]
}

URL_TEMPLATE = "/v1/functions/%s/versions"


class TestFunctionVersion(test_client.TestQinlingClient):

    _error_message = "Test error message."

    def test_list_function_version(self):
        function_id = VERSION_1['function_id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + URL_TEMPLATE % function_id,
            headers={'Content-Type': 'application/json'},
            json=LIST_FUNCTION_VERSIONS_RESP,
            status_code=200
        )
        ret = self.client.function_versions.list(function_id)
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [VERSION_1, VERSION_2],
            [resource.to_dict() for resource in ret])

    def test_create_function_version(self):
        function_id = VERSION_1['function_id']
        request_data = {'description': ''}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + URL_TEMPLATE % function_id,
            headers={'Content-Type': 'application/json'},
            json=VERSION_1,
            status_code=201
        )
        ret = self.client.function_versions.create(function_id)
        self.assertEqual(VERSION_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_function_version_all_options(self):
        function_id = VERSION_1['function_id']
        description = 'This a newly created function version.'
        request_data = {'description': description}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + URL_TEMPLATE % function_id,
            headers={'Content-Type': 'application/json'},
            json=VERSION_1,
            status_code=201
        )
        ret = self.client.function_versions.create(
            function_id, description=description)
        self.assertEqual(VERSION_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_function_version_error(self):
        function_id = VERSION_1['function_id']
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + URL_TEMPLATE % function_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=500
        )
        self.assertRaisesRegex(
            exceptions.HTTPInternalServerError,
            self._error_message,
            self.client.function_versions.create,
            function_id)

    def test_delete_function_version(self):
        function_id = VERSION_1['function_id']
        version = VERSION_1['version_number']
        url = URL_TEMPLATE % function_id + '/%s' % version
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + url,
            status_code=204
        )
        ret = self.client.function_versions.delete(function_id, version)
        self.assertIsNone(ret)

    def test_delete_function_version_error(self):
        function_id = VERSION_1['function_id']
        version = VERSION_1['version_number']
        url = URL_TEMPLATE % function_id + '/%s' % version
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + url,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=403
        )
        self.assertRaisesRegex(
            exceptions.HTTPForbidden,
            self._error_message,
            self.client.function_versions.delete,
            function_id, version
        )

    def test_get_function_version(self):
        function_id = VERSION_2['function_id']
        version = VERSION_2['version_number']
        url = URL_TEMPLATE % function_id + '/%s' % version
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + url,
            headers={'Content-Type': 'application/json'},
            json=VERSION_2,
            status_code=200
        )
        ret = self.client.function_versions.get(function_id, version)
        self.assertEqual(VERSION_2, ret.to_dict())

    def test_get_function_version_error(self):
        function_id = VERSION_2['function_id']
        version = VERSION_2['version_number']
        url = URL_TEMPLATE % function_id + '/%s' % version
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + url,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_versions.get,
            function_id, version
        )

    def test_detach_function_version(self):
        function_id = VERSION_2['function_id']
        version = VERSION_2['version_number']
        url = URL_TEMPLATE % function_id + '/%s/detach' % version
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + url,
            status_code=202
        )
        ret = self.client.function_versions.detach(function_id, version)
        self.assertEqual('', ret.text)
        self.assertEqual(202, ret.status_code)

    def test_detach_function_version_error(self):
        function_id = VERSION_2['function_id']
        version = VERSION_2['version_number']
        url = URL_TEMPLATE % function_id + '/%s/detach' % version
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + url,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_versions.detach,
            function_id, version
        )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
import shutil
import tempfile
import zipfile

from osc_lib.command import command
from osc_lib import utils
from oslo_utils import uuidutils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient import utils as q_utils

MAX_ZIP_SIZE = 50 * 1024 * 1024


def _get_package_file(package_path=None, file_path=None):
    if package_path:
        if not zipfile.is_zipfile(package_path):
            raise exceptions.QinlingClientException(
                'Package %s is not a valid ZIP file.' % package_path
            )

        if os.path.getsize(package_path) > MAX_ZIP_SIZE:
            raise exceptions.QinlingClientException(
                'Package file size must be no more than %sM.' %
                (MAX_ZIP_SIZE / 1024 / 1024)
            )

        return package_path

    elif file_path:
        if not os.path.isfile(file_path):
            raise exceptions.QinlingClientException(
                'File %s not exist.' % file_path
            )

        base_name, extension = os.path.splitext(file_path)
        base_name = os.path.basename(base_name)
        zip_file = os.path.join(
            tempfile.gettempdir(),
            '%s.zip' % base_name
        )

        zf = zipfile.ZipFile(zip_file, mode='w')
        try:
            # Use default compression mode, may change in future.
            zf.write(
                file_path,
                '%s%s' % (base_name, extension),
                compress_type=zipfile.ZIP_STORED
            )
        finally:
            zf.close()

        if os.path.getsize(zip_file) > MAX_ZIP_SIZE:
            raise exceptions.QinlingClientException(
                'Package file size must be no more than %sM.' %
                (MAX_ZIP_SIZE / 1024 / 1024)
            )

        return zip_file


def worker_count(value):
    try:
        value = int(value)
        if value <= 0:
            raise ValueError
    except ValueError:
        raise exceptions.QinlingClientException(
            'Worker count must be a positive integer.'
        )
    return value


class List(base.QinlingLister):
    columns = base.FUNCTION_COLUMNS
    filtered_columns = base.FILTERED_FUNCTION_COLUMNS

    def _get_resources(self, parsed_args):
        client = self.app.client_manager.function_engine

        return client.functions.list(**base.get_filters(parsed_args))


class Create(command.ShowOne):
    columns = base.FUNCTION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            "--runtime",
            help="Runtime ID. Runtime is needed for function of package type "
                 "and swift type, but not for the image type function.",
        )
        parser.add_argument(
            "--name",
            help="Function name.",
        )
        parser.add_argument(
            "--entry",
            help="Function entry in the format of <module_name>.<method_name>"
        )
        protected_group = parser.add_mutually_exclusive_group(required=False)
        protected_group.add_argument(
            "--file",
            metavar="CODE_FILE_PATH",
            help="Code file path."
        )
        protected_group.add_argument(
            "--package",
            metavar="CODE_PACKAGE_PATH",
            help="Code package zip file path."
        )
        parser.add_argument(
            "--container",
            help="Container name in Swift.",
        )
        parser.add_argument(
            "--object",
            help="Object name in Swift.",
        )
        parser.add_argument(
            "--image",
            help="Image name in docker hub.",
        )
        parser.add_argument(
            "--cpu",
            type=q_utils.check_positive,
            help="Limit of cpu resource(unit: millicpu).",
        )
        parser.add_argument(
            "--memory-size",
            type=q_utils.check_positive,
            help="Limit of memory resource(unit: bytes).",
        )
        parser.add_argument(
            "--timeout",
            type=q_utils.check_positive,
            default=5,
            help="The function execution time at which Qinling should "
                 "terminate the function. The default is 5 seconds",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        code_type = None

        if (parsed_args.file or parsed_args.package):
            code_type = 'package'
        elif (parsed_args.container or parsed_args.object):
            code_type = 'swift'
        elif parsed_args.image:
            code_type = 'image'
        else:
            raise exceptions.QinlingClientException(
                'Cannot create function with the parameters given.\nMust '
                'provide required parameters for different type of '
                'functions:\n'
                '    - for package type function, either --file or --package '
                'is required,\n'
                '    - for swift type function, both --container and --object '
                'are required,\n'
                '    - for image type function, --image is required.'
            )

        runtime = parsed_args.runtime
        if runtime and not uuidutils.is_uuid_like(runtime):
            # Try to find the runtime id with name
            runtime = q_utils.find_resource_id_by_name(
                client.runtimes, runtime)

        if code_type == 'package':
            if not runtime:
                raise exceptions.QinlingClientException(
                    'Runtime needs to be specified for package type function.'
                )

            zip_file = _get_package_file(parsed_args.package, parsed_args.file)
            md5sum = q_utils.md5(file=zip_file)
            code = {"source": "package", "md5sum": md5sum}

            with open(zip_file, 'rb') as package:
                function = client.functions.create(
                    name=parsed_args.name,
                    runtime=runtime,
                    code=code,
                    package=package,
                    entry=parsed_args.entry,
                    cpu=parsed_args.cpu,
                    memory_size=parsed_args.memory_size,
                    timeout=parsed_args.timeout
                )

            # Delete zip file the client created
            if parsed_args.file and not parsed_args.package:
                os.remove(zip_file)

        elif code_type == 'swift':
            if not (parsed_args.container and parsed_args.object):
                raise exceptions.QinlingClientException(
                    'Container name and object name need to be specified.'
                )
            if not runtime:
                raise exceptions.QinlingClientException(
                    'Runtime needs to be specified for swift type function.'
                )

            code = {
                "source": "swift",
                "swift": {
                    "container": parsed_args.container,
                    "object": parsed_args.object
                }
            }

            function = client.functions.create(
                name=parsed_args.name,
                runtime=runtime,
                code=code,
                entry=parsed_args.entry,
                cpu=parsed_args.cpu,
                memory_size=parsed_args.memory_size,
                timeout=parsed_args.timeout
            )

        elif code_type == 'image':
            code = {
                "source": "image",
                "image": parsed_args.image
            }

            function = client.functions.create(
                name=parsed_args.name,
                code=code,
                entry=parsed_args.entry,
                cpu=parsed_args.cpu,
                memory_size=parsed_args.memory_size,
                timeout=parsed_args.timeout
            )

        return self.columns, utils.get_item_properties(function, self.columns)


class Delete(base.QinlingDeleter):
    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'function',
            nargs='+',
            metavar='FUNCTION',
            help='Id or name of function(s).'
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        self.delete = client.functions.delete
        self.resource = 'function'

        ids = []
        for function_id in parsed_args.function:
            if not uuidutils.is_uuid_like(function_id):
                function_id = q_utils.find_resource_id_by_name(
                    client.functions, function_id)
            ids.append(function_id)

        self.delete_resources(ids)


class Show(command.ShowOne):
    columns = base.FUNCTION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Show, self).get_parser(prog_name)
        parser.add_argument('function', help='Function ID or name.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        function_id = parsed_args.function
        if not uuidutils.is_uuid_like(function_id):
            function_id = q_utils.find_resource_id_by_name(
                client.functions, function_id)

        function = client.functions.get(function_id)
        return self.columns, utils.get_item_properties(function,
                                                       self.columns)


class Update(command.ShowOne):
    columns = base.FUNCTION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'function',
            help='Function ID or name.'
        )
        parser.add_argument(
            "--name",
            help="Function name."
        )
        parser.add_argument(
            "--description",
            help="Function description."
        )
        parser.add_argument(
            "--entry",
            help="Function entry in the format of <module_name>.<method_name>"
        )

        package_group = parser.add_mutually_exclusive_group()
        package_group.add_argument(
            "--file",
            metavar="CODE_FILE_PATH",
            help="Code file path."
        )
        package_group.add_argument(
            "--package",
            metavar="CODE_PACKAGE_PATH",
            help="Code package zip file path."
        )

        # For swift functions
        parser.add_argument(
            "--container",
            help="Container name in Swift.",
        )
        parser.add_argument(
            "--object",
            help="Object name in Swift.",
        )

        parser.add_argument(
            "--cpu",
            type=q_utils.check_positive,
            help="Limit of cpu resource(unit: millicpu).",
        )
        parser.add_argument(
            "--memory-size",
            type=q_utils.check_positive,
            help="Limit of memory resource(unit: bytes).",
        )
        parser.add_argument(
            "--timeout",
            type=q_utils.check_positive,
            help="The function execution time at which Qinling should "
                 "terminate the function. The default is 5 seconds",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        code = None
        package = None
        zip_file = None

        function_id = parsed_args.function
        if not uuidutils.is_uuid_like(function_id):
            function_id = q_utils.find_resource_id_by_name(
                client.functions, function_id)

        if parsed_args.file or parsed_args.package:
            code = {'source': 'package'}
            zip_file = _get_package_file(parsed_args.package, parsed_args.file)
        elif parsed_args.container or parsed_args.object:
            swift = {}
            if parsed_args.container:
                swift["container"] = parsed_args.container
            if parsed_args.object:
                swift["object"] = parsed_args.object
            code = {'source': 'swift', 'swift': swift}

        if zip_file:
            with open(zip_file, 'rb') as package:
                func = client.functions.update(
                    function_id,
                    code=code,
                    package=package,
                    name=parsed_args.name,
                    description=parsed_args.description,
                    entry=parsed_args.entry,
                    cpu=parsed_args.cpu,
                    memory_size=parsed_args.memory_size,
                    timeout=parsed_args.timeout
                )
        else:
            func = client.functions.update(
                function_id,
                code=code,
                name=parsed_args.name,
                description=parsed_args.description,
                entry=parsed_args.entry,
                cpu=parsed_args.cpu,
                memory_size=parsed_args.memory_size,
                timeout=parsed_args.timeout
            )

        return self.columns, utils.get_item_properties(func, self.columns)


class Detach(command.Command):
    def get_parser(self, prog_name):
        parser = super(Detach, self).get_parser(prog_name)
        parser.add_argument('function', help='Function ID.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        success_msg = "Request to detach function %s has been accepted."
        error_msg = "Unable to detach the specified function."

        try:
            client.functions.detach(parsed_args.function)
            print(success_msg % parsed_args.function)
        except Exception as e:
            print(e)
            raise exceptions.QinlingClientException(error_msg)


class Download(command.Command):
    def get_parser(self, prog_name):
        parser = super(Download, self).get_parser(prog_name)
        parser.add_argument('function', help='Function ID or name.')
        parser.add_argument(
            "-o",
            "--output",
            help="Target file path. If not provided, function ID will be used"
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        function_id = parsed_args.function
        if not uuidutils.is_uuid_like(function_id):
            function_id = q_utils.find_resource_id_by_name(
                client.functions, function_id)

        res = client.functions.get(function_id, download=True)

        cwd = os.getcwd()
        if parsed_args.output:
            if os.path.isabs(parsed_args.output):
                abs_path = parsed_args.output
            else:
                abs_path = os.path.join(cwd, parsed_args.output)
        else:
            abs_path = os.path.join(cwd, "%s.zip" % function_id)

        with open(abs_path, 'wb') as target:
            shutil.copyfileobj(res.raw, target)
        print("Code package downloaded to %s" % (abs_path))


class Scaleup(command.Command):
    def get_parser(self, prog_name):
        parser = super(Scaleup, self).get_parser(prog_name)
        parser.add_argument('function', help='Function ID.')
        parser.add_argument('--count', type=worker_count, default=1,
                            help='Number of workers to scale up.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        success_msg = "Request to scale up function %s has been accepted."
        error_msg = "Unable to scale up the specified function."

        try:
            client.functions.scaleup(parsed_args.function, parsed_args.count)
            print(success_msg % parsed_args.function)
        except Exception as e:
            print(e)
            raise exceptions.QinlingClientException(error_msg)


class Scaledown(command.Command):
    def get_parser(self, prog_name):
        parser = super(Scaledown, self).get_parser(prog_name)
        parser.add_argument('function', help='Function ID.')
        parser.add_argument('--count', type=worker_count, default=1,
                            help='Number of workers to scale down.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        success_msg = "Request to scale down function %s has been accepted."
        error_msg = "Unable to scale down the specified function."

        try:
            client.functions.scaledown(parsed_args.function, parsed_args.count)
            print(success_msg % parsed_args.function)
        except Exception as e:
            print(e)
            raise exceptions.QinlingClientException(error_msg)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
import textwrap

from osc_lib.command import command
from osc_lib import utils
import six

from qinlingclient.common import exceptions
from qinlingclient.i18n import _

RUNTIME_COLUMNS = (
    'id',
    'name',
    'image',
    'status',
    'description',
    'is_public',
    'trusted',
    'project_id',
    'created_at',
    'updated_at'
)
RUNTIME_POOL_COLUMNS = (
    'name',
    'capacity'
)
FUNCTION_COLUMNS = (
    'id',
    'name',
    'description',
    'count',
    'code',
    'runtime_id',
    'entry',
    'project_id',
    'created_at',
    'updated_at',
    'cpu',
    'memory_size',
    'timeout'
)
FILTERED_FUNCTION_COLUMNS = (
    "all_projects",
    "project_id"
)
EXECUTION_COLUMNS = (
    'id',
    'function_alias',
    'function_id',
    'function_version',
    'description',
    'input',
    'result',
    'status',
    'sync',
    'project_id',
    'created_at',
    'updated_at'
)
FILTERED_EXECUTION_COLUMNS = (
    'function_id',
    'project_id',
    'all_projects',
    'status',
    'description',
)
JOB_COLUMNS = (
    'id',
    'name',
    'count',
    'status',
    'function_alias',
    'function_id',
    'function_version',
    'function_input',
    'pattern',
    'first_execution_time',
    'next_execution_time',
    'project_id',
    'created_at',
    'updated_at'
)
FILTERED_JOB_COLUMNS = (
    "all_projects",
    "project_id"
)
WORKER_COLUMNS = (
    'function_id',
    'worker_name'
)
WEBHOOK_COLUMNS = (
    'id',
    'function_alias',
    'function_id',
    'function_version',
    'description',
    'project_id',
    'created_at',
    'updated_at',
    'webhook_url'
)
FILTERED_WEBHOOK_COLUMNS = (
    "all_projects",
    "project_id"
)
FUNCTION_VERSION_COLUMNS = (
    'id',
    'function_id',
    'description',
    'version_number',
    'count',
    'project_id',
    'created_at',
    'updated_at'
)
FUNCTION_ALIAS_COLUMNS = (
    'name',
    'function_id',
    'description',
    'function_version',
    'project_id',
    'created_at',
    'updated_at'
)
FILTERED_FUNCTION_ALIAS_COLUMNS = (
    "all_projects",
    "project_id"
)


@six.add_metaclass(abc.ABCMeta)
class QinlingLister(command.Lister):
    columns = ()
    filtered_columns = ()

    def get_parser(self, prog_name):
        parser = super(QinlingLister, self).get_parser(prog_name)
        parser.add_argument(
            '--filter',
            dest='filters',
            action='append',
            help=_(
                'Filters for resource query that can be repeated. Supported '
                'operands: eq, neq, in, nin, gt, gte, lt, lte, has. '
                'E.g. --filter key="neq:123". The available keys: {0}'
            ).format(self.filtered_columns)
        )

        return parser

    @abc.abstractmethod
    def _get_resources(self, parsed_args):
        """Gets a list of API resources (e.g. using client)."""
        raise NotImplementedError

    def _validate_parsed_args(self, parsed_args):
        # No-op by default.
        pass

    def _headers(self):
        return [c.capitalize() for c in self.columns]

    def take_action(self, parsed_args):
        self._validate_parsed_args(parsed_args)

        ret = self._get_resources(parsed_args)
        if not isinstance(ret, list):
            ret = [ret]

        return (
            self._headers(),
            list(utils.get_item_properties(
                s,
                self.columns,
            ) for s in ret)
        )


class QinlingDeleter(command.Command):
    def delete_resources(self, ids):
        """Delete one or more resources."""
        failure_flag = False
        success_msg = "Request to delete %s %s has been accepted."
        error_msg = "Unable to delete the specified %s(s)."

        for id in ids:
            try:
                self.delete(id)
                print(success_msg % (self.resource, id))
            except Exception as e:
                failure_flag = True
                print(e)

        if failure_flag:
            raise exceptions.QinlingClientException(error_msg % self.resource)


def cut(string, length=25):
    if string and len(string) > length:
        return "%s..." % string[:length]
    else:
        return string


def wrap(string, width=25):
    if string and len(string) > width:
        return textwrap.fill(string, width)
    else:
        return string


def get_filters(parsed_args):
    filters = {}

    if hasattr(parsed_args, 'filters') and parsed_args.filters:
        for f in parsed_args.filters:
            arr = f.split('=')

            if len(arr) != 2:
                raise ValueError('Invalid filter: %s' % f)

            filters[arr[0]] = arr[1]

    return filters



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.common import base


class Job(base.Resource):
    pass


class JobManager(base.Manager):
    resource_class = Job

    def list(self, **kwargs):
        return self._list("/v1/jobs", response_key='jobs')

    def create(self, function_alias=None, function_id=None, function_version=0,
               name=None, first_execution_time=None, pattern=None,
               function_input=None, count=None):
        body = {
            "function_alias": function_alias,
            "function_id": function_id,
            "function_version": function_version,
            "name": name,
            "first_execution_time": first_execution_time,
            "pattern": pattern,
            "function_input": function_input,
            "count": count
        }
        return self._create('/v1/jobs', data=body)

    def delete(self, id):
        self._delete('/v1/jobs/%s' % id)

    def get(self, id):
        return self._get('/v1/jobs/%s' % id)

    def update(self, id, **kwargs):
        return self._update('/v1/jobs/%s' % id, kwargs)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid

from oslo_serialization import jsonutils

from qinlingclient.common import exceptions
from qinlingclient.tests.unit.v1 import test_client

EXECUTION_1 = {'id': str(uuid.uuid4()), 'function_id': str(uuid.uuid4())}
EXECUTION_2 = {'id': str(uuid.uuid4()), 'function_id': str(uuid.uuid4())}
EXECUTION_3 = {'id': str(uuid.uuid4()), 'function_alias': 'alias_1'}

LIST_FUNCTION_EXECUTIONS_RESP = {
    'executions': [EXECUTION_1, EXECUTION_2]
}


class TestFunctionExecution(test_client.TestQinlingClient):

    _error_message = "Test error message."

    def test_list_function_execution(self):
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/executions',
            headers={'Content-Type': 'application/json'},
            json=LIST_FUNCTION_EXECUTIONS_RESP,
            status_code=200
        )
        ret = self.client.function_executions.list()
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [EXECUTION_1, EXECUTION_2],
            [resource.to_dict() for resource in ret])

    def test_list_function_execution_with_params(self):
        self.requests_mock.register_uri(
            'GET',
            (test_client.QINLING_URL +
             '/v1/executions?status=success&sync=True'),
            headers={'Content-Type': 'application/json'},
            json=LIST_FUNCTION_EXECUTIONS_RESP,
            status_code=200
        )
        ret = self.client.function_executions.list(status='success', sync=True)
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [EXECUTION_1, EXECUTION_2],
            [resource.to_dict() for resource in ret])

    def test_create_function_execution(self):
        function_id = EXECUTION_1['function_id']
        request_data = {'function_id': function_id, 'function_version': 0,
                        'function_alias': None, 'sync': True, 'input': None}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/executions',
            headers={'Content-Type': 'application/json'},
            json=EXECUTION_1,
            status_code=201
        )
        ret = self.client.function_executions.create(function_id)
        self.assertEqual(EXECUTION_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_function_execution_all_options(self):
        function_id = EXECUTION_1['function_id']
        function_version = 1
        sync = False
        function_input = '{"name": "Qinling"}'
        request_data = {'function_id': function_id,
                        'function_version': function_version,
                        'function_alias': None,
                        'sync': sync, 'input': function_input}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/executions',
            headers={'Content-Type': 'application/json'},
            json=EXECUTION_1,
            status_code=201
        )
        ret = self.client.function_executions.create(
            function_id, function_version=function_version, sync=sync,
            input=function_input)
        self.assertEqual(EXECUTION_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_function_execution_error(self):
        function_id = EXECUTION_1['function_id']
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/executions',
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=400
        )
        self.assertRaisesRegex(
            exceptions.HTTPBadRequest,
            self._error_message,
            self.client.function_executions.create,
            function_id)

    def test_delete_function_execution(self):
        execution_id = EXECUTION_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/executions/%s' % execution_id,
            status_code=204
        )
        ret = self.client.function_executions.delete(execution_id)
        self.assertIsNone(ret)

    def test_delete_function_execution_error(self):
        execution_id = EXECUTION_1['id']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/executions/%s' % execution_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_executions.delete,
            execution_id
        )

    def test_get_function_execution(self):
        execution_id = EXECUTION_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/executions/%s' % execution_id,
            headers={'Content-Type': 'application/json'},
            json=EXECUTION_2,
            status_code=200
        )
        ret = self.client.function_executions.get(execution_id)
        self.assertEqual(EXECUTION_2, ret.to_dict())

    def test_get_function_execution_error(self):
        execution_id = EXECUTION_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/executions/%s' % execution_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_executions.get,
            execution_id
        )

    def test_get_function_execution_log(self):
        execution_id = EXECUTION_2['id']
        execution_log = 'Preparing...\nRunning...\nDone.'
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/executions/%s/log' % execution_id,
            text=execution_log,
            headers={'Content-Type': 'text/plain'},
            status_code=200
        )
        ret = self.client.function_executions.get_log(execution_id)
        self.assertEqual(execution_log, ret)

    def test_get_function_execution_log_error(self):
        execution_id = EXECUTION_2['id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/executions/%s/log' % execution_id,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_executions.get_log,
            execution_id
        )



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
from unittest import mock

from osc_lib.tests import utils as osc_tests_utils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import job
from qinlingclient.tests.unit.osc.v1 import fakes


class TestJob(fakes.TestQinlingClient):

    def setUp(self):
        super(TestJob, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.JOB_COLUMNS
        self.data = []

        self._jobs = fakes.FakeJob.create_jobs(count=3)
        for j in self._jobs:
            self.data.append((j.id, j.name, j.count, j.status,
                              j.function_alias,
                              j.function_id, j.function_version,
                              j.function_input, j.pattern,
                              j.first_execution_time, j.next_execution_time,
                              j.project_id, j.created_at, j.updated_at))


class TestListJob(TestJob):

    def setUp(self):
        super(TestListJob, self).setUp()
        self.cmd = job.List(self.app, None)

        self.columns = [c.capitalize() for c in base.JOB_COLUMNS]

        self.client.jobs.list = mock.Mock(return_value=self._jobs)

    def test_job_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.list.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_job_list_with_filter(self):
        arglist = ['--filter', 'name=has:job',
                   '--filter', 'status=eq:running']
        verifylist = [
            ('filters', ['name=has:job', 'status=eq:running']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.list.assert_called_once_with(
            name='has:job', status='eq:running'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_job_list_with_invalid_filter(self):
        arglist = ['--filter', 'name']
        verifylist = [
            ('filters', ['name'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            ValueError,
            '^Invalid filter: name$',
            self.cmd.take_action, parsed_args
        )


class TestCreateJob(TestJob):

    def setUp(self):
        super(TestCreateJob, self).setUp()
        self.cmd = job.Create(self.app, None)

    def _create_fake_job(self, attrs=None):
        # Allow to fake different create results
        j = fakes.FakeJob.create_one_job(attrs)
        self.client.jobs.create = mock.Mock(return_value=j)
        data = (j.id, j.name, j.count, j.status,
                j.function_alias,
                j.function_id, j.function_version,
                j.function_input, j.pattern,
                j.first_execution_time, j.next_execution_time,
                j.project_id, j.created_at, j.updated_at)
        return data

    def test_job_create_function_id(self):
        """Create a job with function id."""
        function_id = self._jobs[0].function_id
        attrs = {'function_id': function_id}
        created_data = self._create_fake_job(attrs)

        arglist = ['--function', function_id]
        verifylist = [
            ('function', function_id),
            ('function_version', 0),
            ('function_alias', None),
            ('name', None),
            ('first_execution_time', None),
            ('pattern', None),
            ('function_input', None),
            ('count', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.create.assert_called_once_with(
            **{'function_alias': None,
               'function_id': function_id, 'function_version': 0,
               'name': None,
               'first_execution_time': None,
               'pattern': None,
               'function_input': None,
               'count': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_job_create_function_name(self):
        """Create a job.

        1. use function name to find the function_id,
        2. all optional params are specified.
        """
        function = fakes.FakeFunction.create_one_function()
        function_name = function.name
        function_id = function.id
        job_name = 'FAKE_JOB_NAME'
        count = 3
        function_version = 1
        function_input = '{"JSON_INPUT_KEY": "JSON_INPUT_VALUE"}'
        pattern = '1 * * * *'
        first_execution_time = str(datetime.datetime.utcnow())
        attrs = {'name': job_name,
                 'count': count,
                 'function_id': function_id,
                 'function_version': function_version,
                 'function_input': function_input,
                 'pattern': pattern,
                 'first_execution_time': first_execution_time}
        created_data = self._create_fake_job(attrs)

        # Use to find the function id with its name
        self.client.functions.find.return_value = function

        arglist = ['--function', function_name,
                   '--function-version', str(function_version),
                   '--name', job_name,
                   '--first-execution-time', first_execution_time,
                   '--pattern', pattern,
                   '--function-input', function_input,
                   '--count', str(count)]
        verifylist = [
            ('function', function_name),
            ('function_version', function_version),
            ('function_alias', None),
            ('name', job_name),
            ('first_execution_time', first_execution_time),
            ('pattern', pattern),
            ('function_input', function_input),
            ('count', count),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.create.assert_called_once_with(
            **{'function_alias': None, 'function_id': function_id,
               'function_version': function_version,
               'name': job_name,
               'first_execution_time': first_execution_time,
               'pattern': pattern,
               'function_input': function_input,
               'count': count}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.functions.find.assert_called_once_with(name=function_name)

    def test_job_create_function_alias(self):
        """Create a job with function alias."""
        function_alias = 'fake_alias'
        attrs = {'function_alias': function_alias}
        created_data = self._create_fake_job(attrs)

        arglist = ['--function-alias', function_alias]
        verifylist = [
            ('function', None),
            ('function_version', 0),
            ('function_alias', function_alias),
            ('name', None),
            ('first_execution_time', None),
            ('pattern', None),
            ('function_input', None),
            ('count', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.create.assert_called_once_with(
            **{'function_alias': function_alias,
               'function_id': None, 'function_version': None,
               'name': None,
               'first_execution_time': None,
               'pattern': None,
               'function_input': None,
               'count': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_job_create_version_not_integer(self):
        # function_version should be an integer value
        function_id = self._jobs[0].function_id

        arglist = ['--function', function_id, '--function-version',
                   'NOT_A_INTEGER']
        verifylist = [
            ('function', function_id),
            ('function_version', 0),
            ('name', None),
            ('first_execution_time', None),
            ('pattern', None),
            ('function_input', None),
            ('count', None),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_job_create_count_not_integer(self):
        # count should be an integer value
        function_id = self._jobs[0].function_id

        arglist = ['--function', function_id, '--count', 'NOT_A_INTEGER']
        verifylist = [
            ('function', function_id),
            ('function_version', 0),
            ('name', None),
            ('first_execution_time', None),
            ('pattern', None),
            ('function_input', None),
            ('count', 'NOT_A_INTEGER'),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestDeleteJob(TestJob):

    def setUp(self):
        super(TestDeleteJob, self).setUp()
        self.cmd = job.Delete(self.app, None)
        self.client.jobs.delete = mock.Mock(return_value=None)

    def test_job_delete_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_job_delete_one(self):
        job_id = self._jobs[0].id
        arglist = [job_id]
        verifylist = [('job', [job_id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.jobs.delete.assert_called_once_with(job_id)

    def test_job_delete_multiple(self):
        job_ids = [j.id for j in self._jobs]
        arglist = job_ids
        verifylist = [('job', job_ids)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        calls = [mock.call(j_id) for j_id in job_ids]
        self.assertEqual(len(job_ids), self.client.jobs.delete.call_count)
        self.client.jobs.delete.assert_has_calls(calls)

    def test_job_delete_multiple_exception(self):
        job_ids = [j.id for j in self._jobs]
        arglist = job_ids
        verifylist = [('job', job_ids)]

        self.client.jobs.delete = mock.Mock(side_effect=[
            None, RuntimeError, None
        ])

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to delete the specified job\(s\)\.$',
            self.cmd.take_action, parsed_args)

        # The second deleteion failed, but the third is done normally
        calls = [mock.call(j_id) for j_id in job_ids]
        self.assertEqual(len(job_ids), self.client.jobs.delete.call_count)
        self.client.jobs.delete.assert_has_calls(calls)


class TestShowJob(TestJob):

    def setUp(self):
        super(TestShowJob, self).setUp()
        self.cmd = job.Show(self.app, None)
        self.client.jobs.get = mock.Mock(return_value=self._jobs[0])

    def test_job_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_job_show(self):
        job_id = self._jobs[0].id
        arglist = [job_id]
        verifylist = [('job', job_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.get.assert_called_once_with(job_id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)


class TestUpdateJob(TestJob):

    def setUp(self):
        super(TestUpdateJob, self).setUp()
        self.cmd = job.Update(self.app, None)

    def _update_fake_job(self, attrs=None):
        # Allow to fake different update results
        j = fakes.FakeJob.create_one_job(attrs)
        self.client.jobs.update = mock.Mock(return_value=j)
        data = (j.id, j.name, j.count, j.status,
                j.function_alias,
                j.function_id, j.function_version,
                j.function_input, j.pattern,
                j.first_execution_time, j.next_execution_time,
                j.project_id, j.created_at, j.updated_at)
        return data

    def test_job_update_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_job_update_required_options(self):
        """Update a job.

        Do nothing as only the job_id is specified.
        """
        job_id = self._jobs[0].id
        attrs = {'id': job_id}
        updated_data = self._update_fake_job(attrs)

        arglist = [job_id]
        verifylist = [
            ('id', job_id),
            ('name', None),
            ('status', None),
            ('next_execution_time', None),
            ('pattern', None),
            ('function_input', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.update.assert_called_once_with(
            job_id,
            **{'name': None,
               'status': None,
               'pattern': None,
               'next_execution_time': None,
               'function_input': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_job_update_all_options(self):
        job_id = self._jobs[0].id
        name = 'UPDATED_FAKE_JOB_NAME'
        status = 'paused'
        next_execution_time = str(
            datetime.datetime.utcnow() + datetime.timedelta(0, 3600))
        pattern = '* 1 * * *'
        function_input = '{"JSON_INPUT_KEY": "JSON_INPUT_VALUE"}'
        attrs = {'id': job_id, 'name': name, 'status': status,
                 'next_execution_time': next_execution_time,
                 'pattern': pattern, 'function_input': function_input}
        updated_data = self._update_fake_job(attrs)

        arglist = [job_id, '--name', name, '--status', status,
                   '--next-execution-time', next_execution_time,
                   '--pattern', pattern, '--function-input', function_input]
        verifylist = [
            ('id', job_id),
            ('name', name),
            ('status', status),
            ('next_execution_time', next_execution_time),
            ('pattern', pattern),
            ('function_input', function_input),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.jobs.update.assert_called_once_with(
            job_id,
            **{'name': name,
               'status': status,
               'pattern': pattern,
               'next_execution_time': next_execution_time,
               'function_input': function_input}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_job_update_status_not_in_choices(self):
        job_id = self._jobs[0].id
        arglist = [job_id, '--status', 'NOT_IN_CHOICES']
        verifylist = [
            ('id', job_id),
            ('name', None),
            ('status', 'NOT_IN_CHOICES'),
            ('next_execution_time', None),
            ('pattern', None),
            ('function_input', None),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid

from oslo_serialization import jsonutils

from qinlingclient.common import exceptions
from qinlingclient.tests.unit.v1 import test_client

ALIAS_1 = {'name': 'function_alias_1', 'function_id': str(uuid.uuid4())}
ALIAS_2 = {'name': 'function_alias_2', 'function_id': str(uuid.uuid4())}

LIST_FUNCTION_ALIASES_RESP = {
    'function_aliases': [ALIAS_1, ALIAS_2]
}


class TestFunctionAlias(test_client.TestQinlingClient):

    _error_message = "Test error message."

    def test_list_function_alias(self):
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/aliases',
            headers={'Content-Type': 'application/json'},
            json=LIST_FUNCTION_ALIASES_RESP,
            status_code=200
        )
        ret = self.client.function_aliases.list()
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [ALIAS_1, ALIAS_2],
            [resource.to_dict() for resource in ret])

    def test_create_function_alias(self):
        name = ALIAS_1['name']
        function_id = ALIAS_1['function_id']
        request_data = {'name': name, 'function_id': function_id,
                        'function_version': 0, 'description': ''}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/aliases',
            headers={'Content-Type': 'application/json'},
            json=ALIAS_1,
            status_code=201
        )
        ret = self.client.function_aliases.create(name, function_id)
        self.assertEqual(ALIAS_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_function_alias_all_options(self):
        name = ALIAS_1['name']
        function_id = ALIAS_1['function_id']
        function_version = 1
        description = 'A newly created function alias.'
        request_data = {'name': name, 'function_id': function_id,
                        'function_version': function_version,
                        'description': description}
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/aliases',
            headers={'Content-Type': 'application/json'},
            json=ALIAS_1,
            status_code=201
        )
        ret = self.client.function_aliases.create(
            name, function_id, function_version=function_version,
            description=description
        )
        self.assertEqual(ALIAS_1, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_create_function_alias_error(self):
        name = ALIAS_1['name']
        function_id = ALIAS_1['function_id']
        self.requests_mock.register_uri(
            'POST',
            test_client.QINLING_URL + '/v1/aliases',
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=400
        )
        self.assertRaisesRegex(
            exceptions.HTTPBadRequest,
            self._error_message,
            self.client.function_aliases.create,
            name, function_id)

    def test_delete_function_alias(self):
        name = ALIAS_1['name']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/aliases/%s' % name,
            status_code=204
        )
        ret = self.client.function_aliases.delete(name)
        self.assertIsNone(ret)

    def test_delete_function_alias_error(self):
        name = ALIAS_1['name']
        self.requests_mock.register_uri(
            'DELETE',
            test_client.QINLING_URL + '/v1/aliases/%s' % name,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_aliases.delete,
            name
        )

    def test_get_function_alias(self):
        name = ALIAS_2['name']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/aliases/%s' % name,
            headers={'Content-Type': 'application/json'},
            json=ALIAS_2,
            status_code=200
        )
        ret = self.client.function_aliases.get(name)
        self.assertEqual(ALIAS_2, ret.to_dict())

    def test_get_function_alias_error(self):
        name = ALIAS_2['name']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/aliases/%s' % name,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_aliases.get,
            name
        )

    def test_update_function_alias(self):
        name = ALIAS_2['name']
        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/aliases/%s' % name,
            headers={'Content-Type': 'application/json'},
            json=ALIAS_2,
            status_code=200
        )
        ret = self.client.function_aliases.update(name)
        self.assertEqual(ALIAS_2, ret.to_dict())

    def test_update_function_alias_all_options(self):
        name = ALIAS_2['name']
        function_id = ALIAS_2['function_id']
        function_version = 2
        description = 'An updated function alias.'
        request_data = {'function_id': function_id,
                        'function_version': function_version,
                        'description': description}
        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/aliases/%s' % name,
            headers={'Content-Type': 'application/json'},
            json=ALIAS_2,
            status_code=200
        )
        ret = self.client.function_aliases.update(
            name, function_id=function_id, function_version=function_version,
            description=description
        )
        self.assertEqual(ALIAS_2, ret.to_dict())
        self.assertEqual(jsonutils.dumps(request_data),
                         self.requests_mock.last_request.text)

    def test_update_function_alias_error(self):
        name = ALIAS_2['name']
        self.requests_mock.register_uri(
            'PUT',
            test_client.QINLING_URL + '/v1/aliases/%s' % name,
            text='{"faultstring": "%s"}' % self._error_message,
            headers={'Content-Type': 'application/json'},
            status_code=404
        )
        self.assertRaisesRegex(
            exceptions.HTTPNotFound,
            self._error_message,
            self.client.function_aliases.update,
            name)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from osc_lib.command import command
from osc_lib import utils

from qinlingclient.osc.v1 import base


class List(base.QinlingLister):
    columns = base.RUNTIME_COLUMNS

    def _get_resources(self, parsed_args):
        client = self.app.client_manager.function_engine

        return client.runtimes.list(**base.get_filters(parsed_args))


class Create(command.ShowOne):
    columns = base.RUNTIME_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            "image",
            metavar='IMAGE',
            help="Container image name used by runtime.",
        )
        parser.add_argument(
            "--name",
            help="Runtime name.",
        )
        parser.add_argument(
            "--description",
            help="Runtime description.",
        )
        parser.add_argument(
            "--private",
            dest='is_public',
            action='store_false',
            help="Create private runtime or not, will create public"
                 "runtime if not specified",
        )
        parser.add_argument(
            "--untrusted",
            dest='trusted',
            action='store_false',
            help="Create untrusted runtime or not, will create trusted "
                 "runtime if not specified",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        runtime = client.runtimes.create(
            name=parsed_args.name,
            description=parsed_args.description,
            image=parsed_args.image,
            is_public=parsed_args.is_public,
            trusted=parsed_args.trusted
        )

        return self.columns, utils.get_item_properties(runtime, self.columns)


class Delete(base.QinlingDeleter):
    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'runtime',
            nargs='+',
            metavar='RUNTIME',
            help='Id of runtime(s).'
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        self.delete = client.runtimes.delete
        self.resource = 'runtime'

        self.delete_resources(parsed_args.runtime)


class Show(command.ShowOne):
    columns = base.RUNTIME_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Show, self).get_parser(prog_name)
        parser.add_argument('runtime', help='Runtime ID.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        runtime = client.runtimes.get(parsed_args.runtime)

        return self.columns, utils.get_item_properties(runtime, self.columns)


class Pool(command.ShowOne):
    columns = base.RUNTIME_POOL_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Pool, self).get_parser(prog_name)
        parser.add_argument('runtime', help='Runtime ID.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        pool = client.runtimes.get_pool(parsed_args.runtime)
        return self.columns, utils.get_item_properties(pool,
                                                       self.columns)



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.common import base

URL_TEMPLATE = "/v1/functions/%s/versions"


class FunctionVersion(base.Resource):
    pass


class FunctionVersionManager(base.Manager):
    resource_class = FunctionVersion

    def list(self, function_id):
        url = URL_TEMPLATE % function_id
        return self._list(url, response_key='function_versions')

    def create(self, function_id, description=""):
        url = URL_TEMPLATE % function_id
        request_data = {"description": description}

        return self._create(url, data=request_data)

    def delete(self, function_id, version):
        url = URL_TEMPLATE % function_id + '/%s' % version
        self._delete(url)

    def get(self, function_id, version):
        url = URL_TEMPLATE % function_id + '/%s' % version
        return self._get(url)

    def detach(self, function_id, version):
        return self.http_client.request(
            URL_TEMPLATE % function_id + '/%s/detach' % version,
            'POST',
        )



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from osc_lib.tests import utils as osc_tests_utils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import function_execution
from qinlingclient.tests.unit.osc.v1 import fakes


class TestFunctionExecution(fakes.TestQinlingClient):

    def setUp(self):
        super(TestFunctionExecution, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.EXECUTION_COLUMNS
        self.data = []

        self._executions = fakes.FakeExecution.create_executions(count=3)
        for e in self._executions:
            self.data.append((e.id, e.function_alias, e.function_id,
                              e.function_version, e.description, e.input,
                              e.result, e.status, e.sync, e.project_id,
                              e.created_at, e.updated_at))


class TestListFunctionExecution(TestFunctionExecution):

    def setUp(self):
        super(TestListFunctionExecution, self).setUp()
        self.cmd = function_execution.List(self.app, None)

        self.columns = [c.capitalize() for c in base.EXECUTION_COLUMNS]

        self.client.function_executions.list = mock.Mock(
            return_value=self._executions)

    def test_function_execution_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_executions.list.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_function_execution_list_with_filter(self):
        arglist = ['--filter', 'description=has:execution',
                   '--filter', 'status=eq:success']
        verifylist = [
            ('filters', ['description=has:execution', 'status=eq:success'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_executions.list.assert_called_once_with(
            description='has:execution', status='eq:success'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_function_execution_list_with_invalid_filter(self):
        arglist = ['--filter', 'function_id']
        verifylist = [
            ('filters', ['function_id'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            ValueError,
            '^Invalid filter: function_id$',
            self.cmd.take_action, parsed_args
        )


class TestCreateFunctionExecution(TestFunctionExecution):

    def setUp(self):
        super(TestCreateFunctionExecution, self).setUp()
        self.cmd = function_execution.Create(self.app, None)

    def _create_fake_execution(self, attrs=None):
        # Allow to fake different create results
        e = fakes.FakeExecution.create_one_execution(attrs)
        self.client.function_executions.create = mock.Mock(return_value=e)
        data = (e.id, e.function_alias, e.function_id, e.function_version,
                e.description, e.input, e.result, e.status, e.sync,
                e.project_id, e.created_at, e.updated_at)
        return data

    def test_function_execution_create_function_id(self):
        """Create a function execution with function id."""
        function_id = self._executions[0].function_id
        attrs = {'function_id': function_id}
        created_data = self._create_fake_execution(attrs)

        arglist = ['--function', function_id]
        verifylist = [
            ('function', function_id),
            ('function_version', 0),
            ('function_alias', None),
            ('input', None),
            ('sync', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_executions.create.assert_called_once_with(
            **{'function_alias': None,
               'function_id': function_id,
               'function_version': 0,
               'sync': True,
               'input': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_function_execution_create_function_name(self):
        """Create a function execution.

        1. use function name to find the function_id,
        2. all optional params are specified.
        """
        function = fakes.FakeFunction.create_one_function()
        function_name = function.name
        function_id = function.id
        function_version = 1
        function_input = '{"JSON_INPUT_KEY": "JSON_INPUT_VALUE"}'
        is_sync = False
        attrs = {'function_id': function_id,
                 'function_version': function_version,
                 'input': function_input,
                 'sync': is_sync}
        created_data = self._create_fake_execution(attrs)

        # Use to find the function id with its name
        self.client.functions.find.return_value = function

        arglist = ['--function', function_name, '--function-version',
                   str(function_version), '--input', function_input, '--async']
        verifylist = [
            ('function', function_name),
            ('function_version', function_version),
            ('function_alias', None),
            ('input', function_input),
            ('sync', is_sync),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_executions.create.assert_called_once_with(
            **{'function_alias': None,
               'function_id': function_id,
               'function_version': function_version,
               'sync': is_sync,
               'input': function_input}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.functions.find.assert_called_once_with(name=function_name)

    def test_function_execution_create_function_alias(self):
        """Create a function execution with function alias."""
        function_alias = 'fake_alias'
        function_input = '{"JSON_INPUT_KEY": "JSON_INPUT_VALUE"}'
        is_sync = False
        attrs = {'function_alias': function_alias,
                 'input': function_input,
                 'sync': is_sync}
        created_data = self._create_fake_execution(attrs)

        arglist = ['--function-alias', function_alias,
                   '--input', function_input, '--async']
        verifylist = [
            ('function', None),
            ('function_version', 0),
            ('function_alias', function_alias),
            ('input', function_input),
            ('sync', is_sync),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_executions.create.assert_called_once_with(
            **{'function_alias': function_alias,
               'function_id': None,
               'function_version': None,
               'input': function_input,
               'sync': is_sync}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_function_execution_create_sync_async_mutually_exclusive(self):
        function_id = self._executions[0].function_id
        # --sync and --async are mutually exclusive
        arglist = ['--function', function_id, '--sync', '--async']
        verifylist = [
            ('function', self._executions[0].function_id),
            ('function_version', 0),
            ('input', None),
            ('sync', True),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_execution_create_version_not_integer(self):
        # function_version should be an integer value
        function_id = self._executions[0].function_id
        arglist = ['--function', function_id,
                   '--function-version', 'NOT_A_INTEGER']
        verifylist = [
            ('function', function_id),
            ('function_version', 0),
            ('input', None),
            ('sync', True),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestDeleteFunctionExecution(TestFunctionExecution):

    def setUp(self):
        super(TestDeleteFunctionExecution, self).setUp()
        self.cmd = function_execution.Delete(self.app, None)
        self.client.function_executions.delete = mock.Mock(return_value=None)

    def test_function_execution_delete_no_option(self):
        # Unlike other resources, function_execution.Delete does nothing when
        # no arguments are specified.
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_executions.delete.assert_not_called()

    def test_function_execution_delete_one(self):
        execution_id = self._executions[0].id
        arglist = ['--execution', execution_id]
        verifylist = [('execution', [execution_id]), ('function', None)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_executions.delete.assert_called_once_with(
            execution_id
        )

    def test_function_execution_delete_multiple(self):
        execution_ids = [e.id for e in self._executions]
        arglist = ['--execution'] + execution_ids
        verifylist = [('execution', execution_ids), ('function', None)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        calls = [mock.call(e_id) for e_id in execution_ids]
        self.assertEqual(len(execution_ids),
                         self.client.function_executions.delete.call_count)
        self.client.function_executions.delete.assert_has_calls(calls)

    def test_function_execution_delete_multiple_exception(self):
        execution_ids = [e.id for e in self._executions]
        arglist = ['--execution'] + execution_ids
        verifylist = [('execution', execution_ids), ('function', None)]

        self.client.function_executions.delete = mock.Mock(side_effect=[
            None, RuntimeError, None
        ])

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to delete the specified execution\(s\)\.$',
            self.cmd.take_action, parsed_args)

        # The second deletion failed, but the third is done normally
        calls = [mock.call(e_id) for e_id in execution_ids]
        self.assertEqual(len(execution_ids),
                         self.client.function_executions.delete.call_count)
        self.client.function_executions.delete.assert_has_calls(calls)

    def test_function_execution_delete_by_function_id(self):
        execution_id = self._executions[0].id
        function_id = self._executions[0].function_id

        self.client.function_executions.list.return_value = [
            self._executions[0]
        ]

        arglist = ['--function', function_id]
        verifylist = [('execution', None), ('function', [function_id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_executions.delete.assert_called_once_with(
            execution_id
        )
        self.client.function_executions.list.assert_called_once_with(
            function_id=function_id
        )

    def test_function_execution_delete_execution_function_mutually_exclusive(
            self
    ):
        # --execution and --function are mutually exclusive
        execution_id = self._executions[0].id
        function_id = self._executions[0].function_id

        arglist = ['--execution', execution_id, '--function', function_id]
        verifylist = [('execution', [execution_id]),
                      ('function', [function_id])]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestShowFunctionExecution(TestFunctionExecution):

    def setUp(self):
        super(TestShowFunctionExecution, self).setUp()
        self.cmd = function_execution.Show(self.app, None)
        self.client.function_executions.get = mock.Mock(
            return_value=self._executions[0]
        )

    def test_function_execution_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_execution_show(self):
        execution_id = self._executions[0].id
        arglist = [execution_id]
        verifylist = [('execution', execution_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_executions.get.assert_called_once_with(
            execution_id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)


class TestLogShowFunctionExecution(TestFunctionExecution):

    def setUp(self):
        super(TestLogShowFunctionExecution, self).setUp()
        self.cmd = function_execution.LogShow(self.app, None)
        self.app_stdout_write = mock.Mock()
        self.app.stdout.write = self.app_stdout_write

    def test_function_execution_log_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_execution_log_show(self):
        execution_id = self._executions[0].id
        fake_log = 'This is a fake log of an execution.'

        self.client.function_executions.get_log.return_value = fake_log

        arglist = [execution_id]
        verifylist = [('execution', execution_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_executions.get_log.assert_called_once_with(
            execution_id
        )
        self.app_stdout_write.assert_called_once_with(fake_log)

    def test_function_execution_log_show_empty_log(self):
        execution_id = self._executions[0].id

        self.client.function_executions.get_log.return_value = ''

        arglist = [execution_id]
        verifylist = [('execution', execution_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_executions.get_log.assert_called_once_with(
            execution_id
        )
        self.app_stdout_write.assert_called_once_with("\n")



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from osc_lib.command import command
from osc_lib import utils
from oslo_utils import uuidutils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient import utils as q_utils


class List(base.QinlingLister):
    columns = base.FUNCTION_VERSION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(List, self).get_parser(prog_name)

        parser.add_argument(
            "function_id",
            help="Function ID.",
        )

        return parser

    def _get_resources(self, parsed_args):
        client = self.app.client_manager.function_engine

        return client.function_versions.list(parsed_args.function_id)


class Create(command.ShowOne):
    columns = base.FUNCTION_VERSION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            "function",
            help="Function name or ID.",
        )
        parser.add_argument(
            "--description",
            help="Description for the new version.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        function_id = parsed_args.function
        if not uuidutils.is_uuid_like(function_id):
            # Try to find the function id with name
            function_id = q_utils.find_resource_id_by_name(
                client.functions, function_id)

        version = client.function_versions.create(
            function_id,
            description=parsed_args.description,
        )

        return self.columns, utils.get_item_properties(version, self.columns)


class Delete(base.QinlingDeleter):
    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            "function_id",
            help="Function ID.",
        )
        parser.add_argument(
            "version_number",
            help="Function version.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        client.function_versions.delete(parsed_args.function_id,
                                        parsed_args.version_number)


class Show(command.ShowOne):
    columns = base.FUNCTION_VERSION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Show, self).get_parser(prog_name)

        parser.add_argument(
            "function_id",
            help="Function ID.",
        )
        parser.add_argument(
            "version_number",
            help="Function version.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        version = client.function_versions.get(parsed_args.function_id,
                                               parsed_args.version_number)

        return self.columns, utils.get_item_properties(version, self.columns)


class Detach(command.Command):
    def get_parser(self, prog_name):
        parser = super(Detach, self).get_parser(prog_name)

        parser.add_argument(
            "function_id",
            help="Function ID.",
        )
        parser.add_argument(
            "version_number",
            help="Function version.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        success_msg = "Request to detach function %s(version %s) has been " \
                      "accepted."
        error_msg = "Unable to detach the specified function version."

        try:
            client.function_versions.detach(parsed_args.function_id,
                                            parsed_args.version_number)
            print(
                success_msg %
                (parsed_args.function_id, parsed_args.version_number)
            )
        except Exception as e:
            print(e)
            raise exceptions.QinlingClientException(error_msg)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.common import http
from qinlingclient.v1 import function
from qinlingclient.v1 import function_alias
from qinlingclient.v1 import function_execution
from qinlingclient.v1 import function_version
from qinlingclient.v1 import function_worker
from qinlingclient.v1 import job
from qinlingclient.v1 import runtime
from qinlingclient.v1 import webhook


class Client(object):
    """Client for the Qinling v1 API.

    :param string endpoint: A user-supplied endpoint URL for the service.
    :param string token: Token for authentication.
    :param integer timeout: Allows customization of the timeout for client
                            http requests. (optional)
    """

    def __init__(self, *args, **kwargs):
        """Initialize a new client for the Qinling v1 API."""
        self.http_client = http._construct_http_client(*args, **kwargs)

        self.runtimes = runtime.RuntimeManager(self.http_client)
        self.functions = function.FunctionManager(self.http_client)
        self.function_executions = function_execution.ExecutionManager(
            self.http_client)
        self.jobs = job.JobManager(self.http_client)
        self.function_workers = function_worker.WorkerManager(self.http_client)
        self.webhooks = webhook.WebhookManager(self.http_client)
        self.function_versions = function_version.FunctionVersionManager(
            self.http_client)
        self.function_aliases = function_alias.FunctionAliasManager(
            self.http_client)



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import hashlib


def md5(file=None, content=None):
    hash_md5 = hashlib.md5()

    if file:
        with open(file, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
    elif content:
        hash_md5.update(content)

    return hash_md5.hexdigest()


def find_resource_id_by_name(manager, name):
    # An exception will be raised when resource is not found or multiple
    # resources for the name are found.
    resource = manager.find(name=name)
    return resource.id


def check_positive(value):
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(
            "%s is an invalid positive int value" % value
        )
    return ivalue
